@@ -13,6 +13,7 @@ mod mdcat {
 }
 
 fn gen_completions<P: AsRef<Path>>(out_dir: P) -> Result<()> {
+    use clap::CommandFactory;
     use clap_complete::*;
 
     let completions = out_dir.as_ref().join("completions");
@@ -21,7 +22,7 @@ fn gen_completions<P: AsRef<Path>>(out_dir: P) -> Result<()> {
     for shell in [Shell::Bash, Shell::Zsh, Shell::Fish, Shell::PowerShell] {
         generate_to(
             shell,
-            &mut mdcat::app("80"),
+            &mut mdcat::Args::command(),
             "mdcat",
             out_dir.as_ref().as_os_str(),
         )?;
